/* ainv.h: factored approximate inverse preconditioner.
* This is based on the algorithm developed by Benzi and Tuma. See the following,
* for example, for more information about it:
*
*    M. Benzi, C.D. Meyer, and M. Tuma, "A sparse approximate inverse
*    preconditioner for the conjugate gradient method", SIAM J. Sci. Comput.,
*    17 (1996), pp. 1135--1149.
*
*    M. Benzi and M. Tuma, "A sparse approximate inverse preconditioner for
*    nonsymmetric linear systems"
*
*    M. Benzi, J.K. Cullum, and M. Tuma, "Robust approximate inverse
*    preconditioning for the conjugate gradient method", Los Alamos tech report
*    LA-UR-99-2899, June 1999, submitted for publication.
*
* Please note this a straightforward implementation, with no real attempt
* at optimizing for speed yet.
*
* This code is in the public domain, with no guarantees that it works!
*
*  Robert Bridson, January 7th, 2000.
*/

#ifndef MRAI_AINV_H
#define MRAI_AINV_H

/* AINVsym: do the (stabilized, outer-product) AINV for symmetric matrices.
* n is the dimension of the problem.
* operator is the function for applying the operator, used in the form
*    operator (opdata, lenx, nzx, x, &leny, nzy, y) to apply the operator
*    (defined with data in opdata) to the sparse vector x giving result y,
*    another sparse vector.
* opdata is any data needed to specify the operator.
* droptol1 is used to sparsify outer-product updates
* droptol2 is used to filter the result.
* *pZia, *pZja, and *pZ store the upper triangular factor of the approximate
*    inverse column-by-column, not including the unit diagonal.
* D is a vector for the inverses of the pivots. (L^TDL is the approximate
*    inverse)
*/
int AINVsym (int n, void (*operator) (const void *, int, const int *,
     const double *, int *, int *, double *), const void *opdata,
     double droptol1, double droptol2, int **pZia, int **pZja, double **pZ,
     double **pD);


/* AINVnon: do the (stabilized?, outer-product) AINV for nonsymmetric
* matrices.
* n is the dimension of the problem.
* linop is the function for applying the linear operator, used in the form
*    linop (opdata, lenx, nzx, x, &leny, nzy, y) to apply the operator
*    (defined with data in opdata) to the sparse vector x giving result y,
*    another sparse vector. Both sparse vectors must be in sorted order.
* opdata is any data needed to specify the operator.
* tranop and trandata are the same for the transposed operator.
* droptol1 is used to sparsify outer-product updates
* droptol2 is used to filter the result.
* *pZia, *pZja, and *pZ store the upper triangular factor of the approximate
*    inverse column-by-column, not including the unit diagonal.
* *pWia, *pWja, and *pW store the transpose of the lower triangular factor
*    column-by-column, not including the unit diagonal.
* D is a vector for the inverses of the pivots.
*/
int AINVnon (int n, void (*linop) (const void *, int, const int *,
     const double *, int *, int *, double *), const void *opdata,
     void (*tranop) (const void *, int, const int *, const double *,
     int *, int *, double *), const void *trandata, double droptol1,
     double droptol2, int **pZia, int **pZja, double **pZ,
     int **pWia, int **pWja, double **pW, double **pD);

#endif

